!--------------------
PROGRAM AMBEREXAMPLE1
!--------------------
!
! . This example has been designed to illustrate how data may be written out in a form that
! . can be read by newer versions of the DYNAMO library. AMBER format was chosen because of
! . its relative simplicity although it can only handle coordinate data and data associated
! . with the topology of a system and its MM terms. Other data must be written out in
! . arbitrary, but simple, formats such as those given for the energy and the indices of the
! . fixed atoms at the end of the example.
!
! . Other data that might be useful to save are:
!
!   * Energy terms other than the total energy.
!
!   * The charge and multiplicity of a QM system and the indices of the QM atoms.
!
! . These items can all be saved by adapting the code at the end of the example.
!
! . Finally a warning. It is important that the AMBER topology is written BEFORE the QM atoms
! . have been defined (i.e. by a call to MOPAC_SETUP) as otherwise many MM terms involving
! . QM atoms will be omitted. These are likely to be important for the proper setup of a QM/MM
! . system in newer versions of the DYNAMO library.
!

! . Module declarations.
USE DYNAMO

IMPLICIT NONE

! . Parameters.
REAL(DP), PARAMETER :: DISTANCE = 5.0_DP

! . Scalars.
INTEGER :: I, IUNIT = -1

! . Arrays.
REAL ( KIND = DP ), DIMENSION(1:3,1:1) :: ORIGIN

! . Print out the header.
CALL DYNAMO_HEADER

! . Process the MM file.
CALL MM_FILE_PROCESS ( "protein.opls_bin", "../data/protein.opls" )

! . Construct the system file.
CALL MM_SYSTEM_CONSTRUCT ( "protein.opls_bin", "../data/crambin.seq" )

! . Read in some coordinates for crambin.
CALL COORDINATES_READ ( "../data/crambin.crd" )

! . Center the coordinates.
CALL TO_PRINCIPAL_AXES ( ATMCRD, ATMMAS )

! . Fix atoms in residues that are too far from the origin.
ORIGIN = 0.0_DP
CALL ATOMS_FIX ( .NOT. ATOM_SELECTION_BY_DISTANCE ( TEST = "<", CUTOFF = DISTANCE, POINTS = ORIGIN, INCLUSION = "RESIDUE" ) )

! . Calculate and print an energy.
CALL ENERGY

! . Save the coordinates and topology in AMBER format.
CALL AMBERCRD_WRITE ( "amberexample1.crd" )
CALL AMBERTOP_WRITE ( "amberexample1.top" )

! . Save any remaining data that could be useful for reproducing the calculation.
! . Write out the energy.
IUNIT = NEXT_UNIT ( )
OPEN ( UNIT = IUNIT, FILE = "amberexample1.energy" )
WRITE ( IUNIT, * ) ETOTAL
CLOSE ( UNIT = IUNIT )

! . Write out the indices of the fixed atoms.
IUNIT = NEXT_UNIT ( )
OPEN ( UNIT = IUNIT, FILE = "amberexample1.fixed" )
DO I = 1,NATOMS
    IF ( ATMFIX(I) ) WRITE ( IUNIT, * ) I
END DO
CLOSE ( UNIT = IUNIT )

! . Finish up.
CALL DYNAMO_FOOTER

END PROGRAM AMBEREXAMPLE1
