!----------------
PROGRAM EXAMPLE14
!----------------

! . Module declarations.
USE DYNAMO

IMPLICIT NONE

! . Local parameters.
INTEGER, PARAMETER :: NTRIES = 100, SEED0 = 9917151

! . Local scalars.
INTEGER            :: ITRY, TMPBND
REAL ( KIND = DP ) :: TSTART

! . Local arrays.
REAL ( KIND = DP ), DIMENSION(1:NTRIES) :: PE0, PE1, PE2, PE3

! . Local arrays.
REAL ( KIND = DP ), ALLOCATABLE, DIMENSION(:,:) :: TMPCRD

! . Print out the header.
CALL DYNAMO_HEADER

! . Process the MM file for the cluster.
CALL MM_FILE_PROCESS ( "LJ.opls_bin", "../data/LJ.opls" )

! . Construct the system file for the cluster.
CALL MM_SYSTEM_CONSTRUCT ( "LJ.opls_bin", "../data/LJ13.seq" )

! . Add constraints.
CALL ADD_CONSTRAINTS

! . Allocate the temporary coordinate array.
ALLOCATE ( TMPCRD(1:3,1:NATOMS) )

! . Loop over the number of attempts.
DO ITRY = 1,NTRIES

   ! . Set the number of constraints.
   NBONDS = TMPBND

   ! . Generate the coordinates for the cluster atoms.
   CALL STRUCTURE

   ! . Initialize the coordinates.
   ATMCRD = TMPCRD

   ! . Do a short dynamics at TSTART to remove constraints in the system.
   CALL VELOCITY_ASSIGN ( TSTART, PRINT = .FALSE. )
   CALL DYNAMICS_OPTIONS ( 0.001_DP, 100, 0 )
   CALL VELOCITY_VERLET_DYNAMICS ( TARGET_TEMPERATURE = TSTART, SCALE_FREQUENCY = 1, SCALE_OPTION = "CONSTANT" )

   ! . Calculate the starting energy.
   CALL ENERGY ( PRINT = .FALSE. ) ; PE0(ITRY) = ETOTAL

   ! . Save the coordinates.
   TMPCRD = ATMCRD

   ! . Directly minimize the coordinates (with constraints).
   CALL OPTIMIZE_CONJUGATE_GRADIENT ( GRADIENT_TOLERANCE = 1.0E-4_DP, PRINT_FREQUENCY = 0, STEP_NUMBER = 10000 )

   ! . Remove the constraints and reminimize.
   NBONDS = 0
   CALL OPTIMIZE_CONJUGATE_GRADIENT ( GRADIENT_TOLERANCE = 1.0E-4_DP, PRINT_FREQUENCY = 0, STEP_NUMBER = 10000 )
   CALL ENERGY ( PRINT = .FALSE. ) ; PE1(ITRY) = ETOTAL

   ! . Reintroduce the constraints.
   NBONDS = TMPBND

   ! . Reset the coordinates.
   ATMCRD = TMPCRD

   ! . Cool the system from TSTART to TSTOP.
   CALL VELOCITY_ASSIGN ( TSTART, PRINT = .FALSE. )
   CALL DYNAMICS_OPTIONS ( 0.001_DP, 40000, 0 )
   CALL VELOCITY_VERLET_DYNAMICS ( TARGET_TEMPERATURE = TSTART * EXP ( -10.0_DP ), &
                                   SCALE_FREQUENCY = 10, SCALE_OPTION = "EXPONENTIAL" )

   ! . Remove the constraints and reminimize.
   NBONDS = 0
   CALL ENERGY ( PRINT = .FALSE. ) ; PE2(ITRY) = ETOTAL
   CALL OPTIMIZE_CONJUGATE_GRADIENT ( GRADIENT_TOLERANCE = 1.0E-4_DP, PRINT_FREQUENCY = 0, STEP_NUMBER = 10000 )
   CALL ENERGY ( PRINT = .FALSE. ) ; PE3(ITRY) = ETOTAL

END DO

! . Deallocate the temporary space.
DEALLOCATE ( TMPCRD )

! . Write out the results.
CALL WRITE_RESULTS

! . Finish up.
CALL DYNAMO_FOOTER

!==============================================================================
CONTAINS
!==============================================================================

   !-------------------------
   SUBROUTINE ADD_CONSTRAINTS
   !-------------------------

   ! . Local scalars.
   INTEGER :: I, II, J

   ! . Calculate the number of bond constraints.
   NBONDS = ( NATOMS * ( NATOMS - 1 ) ) / 2

   ! . Deallocate the bond array.
   IF ( ALLOCATED ( BONDS ) ) DEALLOCATE ( BONDS )

   ! . Allocate the bond array.
   ALLOCATE ( BONDS(1:NBONDS) )

   ! . Fill the atom bond indices.
   II = 0
   DO I = 1,(NATOMS-1)
      DO J = (I+1),NATOMS
         II = II + 1
         BONDS(II)%I = I
         BONDS(II)%J = J
      END DO
   END DO

   ! . Fill the remaining bond arrays.
   BONDS(1:NBONDS)%EQ = 2.0_DP
   BONDS(1:NBONDS)%FC = 0.2_DP

   ! . Store the number of bonds.
   TMPBND = NBONDS

   END SUBROUTINE ADD_CONSTRAINTS

   !-------------------
   SUBROUTINE STRUCTURE
   !-------------------

   ! . Local parameters.
   INTEGER,            PARAMETER :: NATTEMPTS = 100
   REAL ( KIND = DP ), PARAMETER :: CLOSE     = 6.25_DP, MAXSIDE = 7.5_DP

   ! . Local scalars.
   INTEGER            :: I, J, N
   REAL ( KIND = DP ) :: SIDE

   ! . Initialize the random number seed.
   CALL RANDOM_INITIALIZE ( SEED0 + ITRY )

   ! . Top of the loop over box coordinate generation.
   10 CONTINUE

   ! . Get the side of the box.
   SIDE = MAXSIDE * ( RANDOM ( ) + 1.0_DP )

   ! . Loop over the atoms.
   DO I = 1,NATOMS

      ! . Initialize the number of attempts to get coordinates.
      N = 0

      ! . Top of the loop over tries.
      20 CONTINUE

      ! . Increment and check N.
      N = N + 1
      IF ( N > NATTEMPTS ) GO TO 10

      ! . Choose coordinates for the atoms.
      TMPCRD(1:3,I) = SIDE * RANDOM_VECTOR ( 3 )

      ! . Verify that the coordinates are not too close to any others.
      DO J = 1,(I-1)
         IF ( SUM ( ( TMPCRD(1:3,I) - TMPCRD(1:3,J) )**2 ) < CLOSE ) GO TO 20
      END DO

   END DO

   ! . Generate the starting and stopping temperatures.
   TSTART = 300.0_DP * ( RANDOM ( ) + 1.0_DP )

   END SUBROUTINE STRUCTURE

   !-----------------------
   SUBROUTINE WRITE_RESULTS
   !-----------------------

   ! . Local scalars.
   INTEGER            :: NMIN1, NMIN3
   REAL ( KIND = DP ) :: EMIN1, EMIN3

   ! . Find the lowest energies and their frequencies.
   EMIN1 = MINVAL ( PE1 ) ; NMIN1 = COUNT ( ABS ( PE1 - EMIN1 ) < 1.0E-2_DP )
   EMIN3 = MINVAL ( PE3 ) ; NMIN3 = COUNT ( ABS ( PE3 - EMIN3 ) < 1.0E-2_DP )

   ! . Write out the header.
   WRITE ( OUTPUT, "(/A)" ) "Optimization Results:"
   WRITE ( OUTPUT, "(/3X,A,6X,A,4X,A,5X,A,8X,A)" ) "Attempt", "Initial Energy", &
                         "Minimized Energy", "Annealed Energy", "Final Energy"

   ! . Write out the results.
   DO ITRY = 1,NTRIES
      WRITE ( OUTPUT, "(I10,4F20.3)" ) ITRY, PE0(ITRY), PE1(ITRY), PE2(ITRY), PE3(ITRY)
   END DO

   ! . Write out the lowest values.
   WRITE ( OUTPUT, "(/A,F20.3,I6)" ) "Lowest minimized energy and frequency = ", EMIN1, NMIN1
   WRITE ( OUTPUT, "(/A,F20.3,I6)" ) "Lowest  annealed energy and frequency = ", EMIN3, NMIN3
   
   END SUBROUTINE WRITE_RESULTS

END PROGRAM EXAMPLE14
