!----------------
PROGRAM EXAMPLE22
!----------------

! . Module declarations.
USE DYNAMO

IMPLICIT NONE

! . Program parameters.
CHARACTER ( LEN = 16 ), PARAMETER :: MCFILE = "Example22.mc"
INTEGER,                PARAMETER :: NSIMULATIONS = 21, SOLUTE = 1
REAL ( KIND = DP ),     PARAMETER :: DLAMBDA = 1.0_DP / ( NSIMULATIONS - 1.0_DP ), T = 300.0_DP

! . Program scalars.
INTEGER            :: ISIM
REAL ( KIND = DP ) :: LAMBDA

! . Program arrays.
REAL ( KIND = DP ), DIMENSION(1:NSIMULATIONS) :: DGIJ = 0.0_DP, DGJI = 0.0_DP

! . Scalars for the parameter subroutines.
INTEGER :: N, START, STOP

! . Arrays for the parameter subroutines.
REAL ( KIND = DP ), ALLOCATABLE, DIMENSION(:) :: CHARGE, EPSILOX, SIGMA

! . Print out the header.
CALL DYNAMO_HEADER

! . Initialize the random number seed.
CALL RANDOM_INITIALIZE ( 737171 )

! . Set the Monte Carlo energy options.
CALL MONTE_CARLO_ENERGY_OPTIONS ( CUTOFF = 8.5_DP, SMOOTH = 0.5_DP )

! . Read in the system file.
CALL MM_SYSTEM_CONSTRUCT ( "solvent.opls_bin", "../data/h2o_box216.seq" )

! . Read in the starting equilibrated coordinates.
CALL COORDINATES_READ ( "../data/h2o_box216.crd" )

! . Save the molecule parameters.
CALL PARAMETERS_SAVE

! . Loop over the number of simulations.
DO ISIM = 1,NSIMULATIONS

   ! . Set the value of LAMBDA.
   LAMBDA = REAL ( ( NSIMULATIONS - ISIM ), DP ) * DLAMBDA

   ! . Reset the parameter values.
   CALL PARAMETERS_SET ( LAMBDA, .TRUE. )

   ! . Equilibrate the system at the new values.
   CALL MONTE_CARLO ( 10, 100000, TEMPERATURE = T )

   ! . Collect data.
   CALL MONTE_CARLO ( 20, 100000, TEMPERATURE = T, SAVE_FREQUENCY = 100, FILE = MCFILE )

   ! . Calculate the averages required from the trajectory.
   CALL CALCULATE_FREE_ENERGIES

END DO

! . Write out a summary of the results.
WRITE ( OUTPUT, "(/23('-'),A,23('-'))" ) " Calculated Free Energies "
WRITE ( OUTPUT, "(2(4X,A),2(7X,A),4X,A)" ) "Lambda I", "Lambda J", "dG (I->J)", "dG (I<-J)", "dG (average)"
DO ISIM = 1,(NSIMULATIONS-1)
   LAMBDA = REAL ( ( NSIMULATIONS - ISIM ), DP ) * DLAMBDA
   WRITE ( OUTPUT, "(2F12.4,3G16.4)" ) LAMBDA, LAMBDA - DLAMBDA, DGIJ(ISIM), DGJI(ISIM), &
                                                      0.5_DP * ( DGJI(ISIM) - DGIJ(ISIM) ) 
END DO
WRITE ( OUTPUT, "(A,19X,3G16.4)" ) "Total", SUM ( DGIJ ), SUM ( DGJI ), 0.5_DP * SUM ( DGJI - DGIJ ) 
WRITE ( OUTPUT, "(72('-'))" )

! . Save the final configuration.
CALL COORDINATES_WRITE ( "h2o_box216.crd22" )

! . Free the parameter arrays.
CALL PARAMETERS_FREE

! . Finish up.
CALL DYNAMO_FOOTER

!===============================================================================
CONTAINS
!===============================================================================

   !---------------------------------
   SUBROUTINE CALCULATE_FREE_ENERGIES
   !---------------------------------

   ! . Local parameters.
   REAL ( KIND = DP ), PARAMETER :: RT = R * T

   ! . Local scalars.
   INTEGER            :: IFRAME
   REAL ( KIND = DP ) :: EI, EJ, GB, GF

   ! . Type definitions.
   TYPE(DCD_TYPE) :: MCTRAJECTORY

   ! . Initialize the trajectory.
   CALL DCD_INITIALIZE ( MCTRAJECTORY )

   ! . Activate the trajectory for reading.
   CALL DCD_ACTIVATE_READ ( MCFILE, MCTRAJECTORY )

   ! . Initialize the accumulators.
   GB = 0.0_DP
   GF = 0.0_DP

   ! . Loop over the frames.
   DO IFRAME = 1,MCTRAJECTORY%NFRAMES

      ! . Read the next frame.
      CALL DCD_READ ( MCTRAJECTORY, ATMCRD, BOXL )

      ! . Calculate the energy at i.
      CALL PARAMETERS_SET ( LAMBDA, .FALSE. )
      EI = MONTE_CARLO_ENERGY_ONE ( SOLUTE )

      ! . Calculate the energy at i-1.
      IF ( ISIM > 1 ) THEN
         CALL PARAMETERS_SET ( LAMBDA + DLAMBDA, .FALSE. )
         EJ = MONTE_CARLO_ENERGY_ONE ( SOLUTE ) 
         GB = GB + EXP ( - ( EJ - EI ) / RT )    
      END IF

      ! . Calculate the energy at i+1.
      IF ( ISIM < NSIMULATIONS ) THEN
         CALL PARAMETERS_SET ( LAMBDA - DLAMBDA, .FALSE. )
         EJ = MONTE_CARLO_ENERGY_ONE ( SOLUTE )
         GF = GF + EXP ( - ( EJ - EI ) / RT )    
      END IF

   END DO

   ! . Scale the values.
   GB = GB / REAL ( MCTRAJECTORY%NFRAMES, DP ) 
   GF = GF / REAL ( MCTRAJECTORY%NFRAMES, DP )

   ! . Calculate the values of the free energies for the window.
   IF ( ISIM >            1 ) DGJI(ISIM-1) = - RT * LOG ( GB )
   IF ( ISIM < NSIMULATIONS ) DGIJ(ISIM)   = - RT * LOG ( GF )

   ! . Deactivate the trajectory.
   CALL DCD_DEACTIVATE ( MCTRAJECTORY )

   END SUBROUTINE CALCULATE_FREE_ENERGIES

   !-------------------------
   SUBROUTINE PARAMETERS_FREE
   !-------------------------

   ! . Deallocate the parameter arrays.
   DEALLOCATE ( CHARGE, EPSILOX, SIGMA )

   END SUBROUTINE PARAMETERS_FREE

   !-------------------------
   SUBROUTINE PARAMETERS_SAVE
   !-------------------------

   ! . Get the atom indices for the molecule.
   START = RESIND(SOLUTE)+1
   STOP  = RESIND(SOLUTE+1)
   N     = RESIND(SOLUTE+1) - RESIND(SOLUTE)

   ! . Allocate the parameter arrays.
   ALLOCATE ( CHARGE(1:N), EPSILOX(1:N), SIGMA(1:N) )

   ! . Save the charges.
   CHARGE = ATMCHG(START:STOP)

   ! . Save the LJ epsilon values.
   EPSILOX = ( ATMEPS(START:STOP) / 2.0_DP )**2

   ! . Save the LJ sigma values.
   SIGMA = ATMSIG(START:STOP)**2

   END SUBROUTINE PARAMETERS_SAVE

   !-------------------------------------------
   SUBROUTINE PARAMETERS_SET ( LAMBDA, QPRINT )
   !-------------------------------------------

   ! . Scalar arguments.
   LOGICAL,            INTENT(IN) :: QPRINT
   REAL ( KIND = DP ), INTENT(IN) :: LAMBDA

   ! . Local scalars.
   INTEGER :: I

   ! . Local arrays.
   REAL ( KIND = DP ), DIMENSION(1:N) :: CHGNEW, EPSNEW, SIGNEW

   ! . Calculate the new charges.
   CHGNEW = LAMBDA * CHARGE

   ! . Calculate the new LJ epsilon values.
   EPSNEW = EPSILOX

   ! . Calculate the new LJ sigma values.
   SIGNEW = SIGMA

   ! . Set the values of the parameters.
   ATMCHG(START:STOP) = CHGNEW
   ATMEPS(START:STOP) = 2.0_DP * SQRT ( EPSNEW )
   ATMSIG(START:STOP) = SQRT ( SIGNEW )

   ! . Print the values if necessary.
   IF ( QPRINT ) THEN
      WRITE ( OUTPUT, "(/17('-'),A,F5.2,A,17('-'))" ) " Parameter Values at Lambda = ", LAMBDA, " "
      WRITE ( OUTPUT, "(6X,A,14X,A,13X,A,15X,A)" ) "Atom", "Charge", "Epsilon", "Sigma"
      WRITE ( OUTPUT, "(I10,3G20.10)" ) ( I, CHGNEW(I), EPSNEW(I), SIGNEW(I), I = 1,N )
      WRITE ( OUTPUT, "(70('-'))" )
   END IF

   END SUBROUTINE PARAMETERS_SET

END PROGRAM EXAMPLE22









