!----------------
PROGRAM EXAMPLE24
!----------------

! . Module declarations.
USE DYNAMO

IMPLICIT NONE

! . Program parameters.
INTEGER,            PARAMETER :: NCHANGES = 10
REAL ( KIND = DP ), PARAMETER :: BIGSIZE  = 10.0_DP, DENSITY = 996.0_DP ! kg m^-3.

! . Program scalars.
INTEGER            :: ICHANGE
REAL ( KIND = DP ) :: NEWSIDE, REDUCE, SIDE, TARGET, VOLUME

! . Program arrays.
REAL ( KIND = DP ), ALLOCATABLE, DIMENSION(:,:) :: COORDINATES

! . Print out the header.
CALL DYNAMO_HEADER

! . Define the atom, sequence and Z-matrix data structures.
CALL ZMATRIX_DEFINE ( "../data/h2o.zmatrix" )

! . Calculate the atomic coordinates.
CALL ZMATRIX_BUILD

! . Print the coordinates.
CALL COORDINATES_WRITE

! . Save the coordinates of the molecule.
ALLOCATE ( COORDINATES(1:3,1:NATOMS) ) ; COORDINATES = ATMCRD

! . Read in the system file for the solvent box.
CALL MM_SYSTEM_CONSTRUCT ( "solvent.opls_bin", "../data/h2o_box216.seq" )

! . Calculate the volume of a box that gives the appropriate density (Angstroms^3).
VOLUME = ( SUM ( ATMMAS(1:NATOMS) ) / DENSITY ) * ( AMU_TO_KG * 1.0E+30_DP )

! . Calculate the dimension of a cubic box of the same volume (Angstroms).
TARGET = EXP ( LOG ( VOLUME ) / 3.0_DP )

! . Start off with a very large box.
SIDE = BIGSIZE * TARGET

! . Calculate the reduction factor.
REDUCE = EXP ( - LOG ( BIGSIZE ) / REAL ( NCHANGES, DP ) )

! . Assign a value to the box size.
CALL SYMMETRY_CUBIC_BOX ( SIDE )

! . Generate coordinates for each molecule.
CALL SET_INITIAL_COORDINATES

! . Define the options for the Monte Carlo modules.
CALL MONTE_CARLO_ENERGY_OPTIONS ( CUTOFF = 8.5_DP, SMOOTH = 0.5_DP )

! . Do a Monte Carlo simulation (at constant volume).
CALL MONTE_CARLO ( 10, 10 * NRESID, VOLUME_FREQUENCY = 0 )

! . Gradually reduce the size of the box to the desired value.
DO ICHANGE = 1,NCHANGES

   ! . Calculate and set the new box length.
   NEWSIDE = REDUCE * SIDE
   CALL SYMMETRY_CUBIC_BOX ( NEWSIDE )

   ! . Scale the coordinates.
   CALL SET_INTERMEDIATE_COORDINATES

   ! . Do a Monte Carlo simulation (at constant volume).
   CALL MONTE_CARLO ( 10, 10 * NRESID, VOLUME_FREQUENCY = 0 )

   ! . Reset SIDE.
   SIDE = NEWSIDE

END DO

! . Do a final MC calculation for a fuller equilibration.
CALL MONTE_CARLO ( 10, 100000 )

! . Save the coordinates.
CALL COORDINATES_WRITE ( "water_box216.crd" )

! . Deallocate the temporary arrays.
DEALLOCATE ( COORDINATES )

! . Finish up.
CALL DYNAMO_FOOTER

!===============================================================================
CONTAINS
!===============================================================================

   !---------------------------------
   SUBROUTINE SET_INITIAL_COORDINATES
   !---------------------------------

   ! . Local scalars.
   INTEGER :: I, START, STOP

   ! . Loop over the residues.
   DO I = 1,NRESID

      ! . Set the initial values of the coordinates.
      START = RESIND(I)+1
      STOP  = RESIND(I+1)
      ATMCRD(1:3,START:STOP) = COORDINATES

      ! . Randomly translate the molecule within the box.
      CALL TRANSLATE ( ATMCRD(1:3,START:STOP), SIDE * ( RANDOM_VECTOR ( 3 ) - 0.5_DP ) )

   END DO

   END SUBROUTINE SET_INITIAL_COORDINATES

   !--------------------------------------
   SUBROUTINE SET_INTERMEDIATE_COORDINATES
   !--------------------------------------

   ! . Local scalars.
   INTEGER            :: I, START, STOP
   REAL ( KIND = DP ) :: SCALE

   ! . Local arrays.
   REAL ( KIND = DP ), DIMENSION(1:3) :: DR

   ! . Calculate the scale factor for changing the coordinates of the molecular centers.
   SCALE = NEWSIDE / SIDE

   ! . Translate the coordinates of the atoms in each molecule.
   DO I = 1,NRESID

      ! . Find the atom indices for the molecule.
      START = RESIND(I)+1
      STOP  = RESIND(I+1)

      ! . Find the translation.
      DR = ( SCALE - 1.0_DP ) * CENTER ( ATMCRD(1:3,START:STOP), ATMMAS(START:STOP) )

      ! . Translate the molecule by the required amount.
      CALL TRANSLATE ( ATMCRD(1:3,START:STOP), DR )

   END DO

   END SUBROUTINE SET_INTERMEDIATE_COORDINATES

END PROGRAM EXAMPLE24



