!===============================================================================
!
!        fDynamo v2.2 - a program for performing molecular simulations.
!                    Copyright (C) 2005-2007 Martin J. Field
!
!===============================================================================
!
!       This program is free software; you can redistribute it and/or     
!       modify it under the terms of the GNU General Public License       
!       as published by the Free Software Foundation; either version 2    
!       of the License, or (at your option) any later version.            
!
!       This program is distributed in the hope that it will be useful,   
!       but WITHOUT ANY WARRANTY; without even the implied warranty of    
!       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     
!       GNU General Public License for more details.                      
!
!       You should have received a copy of the GNU General Public License 
!       along with this program; if not, write to the Free Software       
!       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,        
!       MA  02110-1301, USA.                                              
!
!===============================================================================
!
!  Email: martin.field@ibs.fr
!  WWWs:  http://www.ibs.fr and http://www.pdynamo.org
!
!===============================================================================
!                           The Energy Module (MOPAC)
!===============================================================================
!
! . MM energy scalars (all kJ mol^-1):
!
!   EANGLE                         The angle energy.
!   EBOND                          The bond energy.
!   EDIHEDRAL                      The dihedral energy.
!   EELECT                         The electrostatic energy.
!   EIMPROPER                      The improper energy.
!   ELJ                            The Lennard-Jones energy.
!
! . Remaining energy scalars:
!
!   ECONSTRAINT                    The constraint energy (kJ mol^-1).
!   EPI                            The harmonic path integral energy (kJ mol^-1).
!   EQM                            The QM energy (kJ mol^-1).
!   ETOTAL                         The total potential energy (kJ mol^-1).
!   GRMS                           The RMS gradient (kJ mol^1 A^-1).
!   VIRIAL                         The internal virial (multiplied by -3).
!
! . Arrays:
!
!   ATMDER                         The first derivatives.
!   ATMHES                         The second derivative matrix.
!
! . Subroutines:
!
!   ENERGY                         Calculate the potential energy.
!   GRADIENT                       Calculate the energy and the first
!                                  derivatives.
!   HESSIAN                        Calculate the energy and the first and second
!                                  derivatives.
!
!   ENERGY_INITIALIZE              Initialize the energy module.
!   ENERGY_PRINT                   Print the energy terms.
!
! . Notes:
!
!   The VIRIAL calculation is only correct if there are no fixed or quantum
!   atoms. This may be changed in future versions although in practice the
!   calculation of the virial with fixed atoms is not particularly useful.
!
!   ATMDER is always dimensioned to (1:3,1:NATOMS) even if there are fixed
!   atoms (this may change). ATMHES, however, has dimensions (N*(N+1))/2
!   where N = 3 * NFREE.
!
!   In a path integral calculation, the energy values printed are averages
!   for each term over the different polymers.
!
!===============================================================================
MODULE POTENTIAL_ENERGY

! . Module declarations.
USE DEFINITIONS,        ONLY : DP
USE PRINTING,           ONLY : PRINT_LINE, PRINT_SUMMARY_ELEMENT, PRINT_SUMMARY_ENDLINE, PRINT_SUMMARY_OPTIONS, &
                               PRINT_SUMMARY_START, PRINT_SUMMARY_STOP

USE ATOMS,              ONLY : ATMCRD, ATMFIX, NATOMS, NATOMSMM, NATOMSQM, NFIXED, NFREE
USE CONSTRAINT,         ONLY : ENERGY_CONSTRAINT
USE ENERGY_COVALENT,    ONLY : ENERGY_ANGLE, ENERGY_BOND, ENERGY_DIHEDRAL, ENERGY_IMPROPER
USE ENERGY_NON_BONDING, ONLY : ENERGY_NON_BONDING_CALCULATE
USE QUANTUM_ENERGY,     ONLY : ENERGY_QUANTUM

IMPLICIT NONE
PUBLIC
#ifndef PGPC
SAVE
#endif

! . Module scalars.
REAL ( KIND = DP ) :: EANGLE      = 0.0_DP, EBOND  = 0.0_DP, &
                      EDIHEDRAL   = 0.0_DP, EELECT = 0.0_DP, &
                      EIMPROPER   = 0.0_DP, ELJ    = 0.0_DP, &
                      ECONSTRAINT = 0.0_DP, EPI    = 0.0_DP, &
		      EQM         = 0.0_DP, ETOTAL = 0.0_DP, &
		      GRMS        = 0.0_DP, VIRIAL = 0.0_DP

! . Module arrays.
REAL ( KIND = DP ), ALLOCATABLE, DIMENSION(:,:) :: ATMDER
REAL ( KIND = DP ), ALLOCATABLE, DIMENSION(:)   :: ATMHES

!===============================================================================
CONTAINS
!===============================================================================

   !--------------------------
   SUBROUTINE ENERGY ( PRINT )
   !--------------------------

   ! . Scalar arguments.
   LOGICAL, INTENT(IN), OPTIONAL :: PRINT

   ! . Local scalars.
   LOGICAL :: QPRINT

   ! . Reset all the energy module variables.
   CALL ENERGY_INITIALIZE

   ! . Return if there are no atoms.
   IF ( NATOMS <= 0 ) RETURN

   ! . Check the print flag.
   IF ( PRESENT ( PRINT ) ) THEN
      QPRINT = PRINT
   ELSE
      QPRINT = .TRUE.
   END IF

   ! . Process the different MM energy terms.
   IF ( NATOMSMM > 0 ) THEN
      CALL ENERGY_BOND                  ( EBOND,       VIRIAL )
      CALL ENERGY_ANGLE                 ( EANGLE              )
      CALL ENERGY_DIHEDRAL              ( EDIHEDRAL           )
      CALL ENERGY_IMPROPER              ( EIMPROPER           )
      CALL ENERGY_NON_BONDING_CALCULATE ( EELECT, ELJ, VIRIAL )
   END IF

   ! . Do a QM calculation.
   IF ( NATOMSQM > 0 ) CALL ENERGY_QUANTUM ( EQM, EPI, VIRIAL, PRINT = QPRINT )

   ! . Calculate the constraint energy.
   CALL ENERGY_CONSTRAINT ( ECONSTRAINT, VIRIAL )

   ! . Sum up the total energy.
   ETOTAL = EANGLE + EBOND + EDIHEDRAL + EIMPROPER + EELECT + ELJ + EQM + EPI + ECONSTRAINT

   ! . Print out the energy if required.
   IF ( QPRINT ) CALL ENERGY_PRINT

   END SUBROUTINE ENERGY

   !----------------------------
   SUBROUTINE GRADIENT ( PRINT )
   !----------------------------

   ! . Scalar arguments.
   LOGICAL, INTENT(IN), OPTIONAL :: PRINT

   ! . Local scalars.
   INTEGER :: IATOM
   LOGICAL :: QPRINT

   ! . Reset all the energy module variables.
   CALL ENERGY_INITIALIZE

   ! . Return if there are no free atoms.
   IF ( NFREE <= 0 ) RETURN

   ! . Check the print flag.
   IF ( PRESENT ( PRINT ) ) THEN
      QPRINT = PRINT
   ELSE
      QPRINT = .TRUE.
   END IF

   ! . Allocate and initialize the gradient array.
   ALLOCATE ( ATMDER(1:3,1:NATOMS) ) ; ATMDER = 0.0_DP

   ! . Process the different MM energy terms.
   IF ( NATOMSMM > 0 ) THEN
      CALL ENERGY_BOND                  ( EBOND,       VIRIAL, ATMDER )
      CALL ENERGY_ANGLE                 ( EANGLE,              ATMDER )
      CALL ENERGY_DIHEDRAL              ( EDIHEDRAL,           ATMDER )
      CALL ENERGY_IMPROPER              ( EIMPROPER,           ATMDER )
      CALL ENERGY_NON_BONDING_CALCULATE ( EELECT, ELJ, VIRIAL, ATMDER )
   END IF

   ! . Do a QM calculation.
   IF ( NATOMSQM > 0 ) CALL ENERGY_QUANTUM ( EQM, EPI, VIRIAL, GRADIENT = ATMDER, PRINT = QPRINT )

   ! . Calculate the constraint energy.
   CALL ENERGY_CONSTRAINT ( ECONSTRAINT, VIRIAL, ATMDER )

   ! . Sum up the total energy.
   ETOTAL = EANGLE + EBOND + EDIHEDRAL + EIMPROPER + EELECT + ELJ + EQM + EPI + ECONSTRAINT

   ! . Zero out the derivatives of fixed atoms.
   IF ( NFIXED > 0 ) THEN
      DO IATOM = 1,NATOMS
         IF ( ATMFIX(IATOM) ) ATMDER(1:3,IATOM) = 0.0_DP
      END DO
   END IF

   ! . Print out the energy if required.
   IF ( QPRINT ) CALL ENERGY_PRINT

   END SUBROUTINE GRADIENT

   !---------------------------
   SUBROUTINE HESSIAN ( PRINT )
   !---------------------------

   ! . Scalar arguments.
   LOGICAL, INTENT(IN), OPTIONAL :: PRINT

   ! . Local scalars.
   INTEGER :: IATOM
   LOGICAL :: QPRINT

   ! . Reset all the energy module variables.
   CALL ENERGY_INITIALIZE

   ! . Return if there are no free atoms.
   IF ( NFREE <= 0 ) RETURN

   ! . Check the print flag.
   IF ( PRESENT ( PRINT ) ) THEN
      QPRINT = PRINT
   ELSE
      QPRINT = .TRUE.
   END IF

   ! . Allocate and initialize the gradient array.
   ALLOCATE ( ATMDER(1:3,1:NATOMS) ) ; ATMDER = 0.0_DP

   ! . Allocate and initialize the Hessian array.
   ALLOCATE ( ATMHES(1:(3*NFREE*(3*NFREE+1))/2) ) ; ATMHES = 0.0_DP

   ! . Process the different energy terms.
   IF ( NATOMSMM > 0 ) THEN
      CALL ENERGY_BOND                  ( EBOND,       VIRIAL, ATMDER, ATMHES )
      CALL ENERGY_ANGLE                 ( EANGLE,              ATMDER, ATMHES )
      CALL ENERGY_DIHEDRAL              ( EDIHEDRAL,           ATMDER, ATMHES )
      CALL ENERGY_IMPROPER              ( EIMPROPER,           ATMDER, ATMHES )
      CALL ENERGY_NON_BONDING_CALCULATE ( EELECT, ELJ, VIRIAL, ATMDER, ATMHES )
   END IF

   ! . Do a QM calculation.
   IF ( NATOMSQM > 0 ) CALL ENERGY_QUANTUM ( EQM, EPI, VIRIAL, ATMDER, ATMHES, QPRINT )

   ! . Calculate the constraint energy.
   CALL ENERGY_CONSTRAINT ( ECONSTRAINT, VIRIAL, ATMDER, ATMHES )

   ! . Sum up the total energy.
   ETOTAL = EANGLE + EBOND + EDIHEDRAL + EIMPROPER + EELECT + ELJ + EQM + EPI + ECONSTRAINT

   ! . Zero out the derivatives of fixed atoms.
   IF ( NFIXED > 0 ) THEN
      DO IATOM = 1,NATOMS
         IF ( ATMFIX(IATOM) ) ATMDER(1:3,IATOM) = 0.0_DP
      END DO
   END IF

   ! . Print out the energy if required.
   IF ( QPRINT ) CALL ENERGY_PRINT

   END SUBROUTINE HESSIAN

   !---------------------------
   SUBROUTINE ENERGY_INITIALIZE
   !---------------------------

   ! . Initialize the MM energy terms.
   EBOND     = 0.0_DP
   EANGLE    = 0.0_DP
   EDIHEDRAL = 0.0_DP
   EIMPROPER = 0.0_DP
   EELECT    = 0.0_DP
   ELJ       = 0.0_DP

   ! . Initialize the remaining energy terms.
   ECONSTRAINT = 0.0_DP
   EPI         = 0.0_DP
   EQM         = 0.0_DP
   ETOTAL      = 0.0_DP
   GRMS        = 0.0_DP
   VIRIAL      = 0.0_DP

   ! . Deallocate the derivative arrays.
   IF ( ALLOCATED ( ATMDER ) ) DEALLOCATE ( ATMDER )
   IF ( ALLOCATED ( ATMHES ) ) DEALLOCATE ( ATMHES )

   END SUBROUTINE ENERGY_INITIALIZE

   !----------------------
   SUBROUTINE ENERGY_PRINT
   !----------------------

   ! . Write out the header.
   CALL PRINT_SUMMARY_OPTIONS ( HEADER_COLOR = "#BB00BB", VARIABLEWIDTH = 16 )
   CALL PRINT_SUMMARY_START ( "Potential Energy Terms" )

   ! . Write out the total energy and the RMS gradient.
   IF ( ALLOCATED ( ATMDER ) ) THEN

      ! . Calculate the RMS gradient.
      GRMS = SQRT ( SUM ( ATMDER * ATMDER ) / REAL ( 3*NFREE, DP ) )

      ! . Write out the energy.
      WRITE ( PRINT_LINE, "(F16.4)" ) ETOTAL ; CALL PRINT_SUMMARY_ELEMENT ( "Potential Energy" )
      WRITE ( PRINT_LINE, "(F16.4)" ) GRMS   ; CALL PRINT_SUMMARY_ELEMENT ( "RMS Gradient"     )

   ! . Write out the total energy.
   ELSE
      WRITE ( PRINT_LINE, "(F16.4)" ) ETOTAL ; CALL PRINT_SUMMARY_ELEMENT ( "Potential Energy" )
      CALL PRINT_SUMMARY_ENDLINE
   END IF

   ! . Write out the QM energy terms.
   IF ( NATOMSQM > 0 ) THEN
      WRITE ( PRINT_LINE, "(F16.4)" ) EQM ; CALL PRINT_SUMMARY_ELEMENT ( "QM Energy" )
      WRITE ( PRINT_LINE, "(F16.4)" ) EPI ; CALL PRINT_SUMMARY_ELEMENT ( "PI Energy" )
   END IF
   
   ! . Write out the MM energy terms.
   IF ( NATOMSMM > 0 ) THEN
      WRITE ( PRINT_LINE, "(F16.4)" ) EBOND     ; CALL PRINT_SUMMARY_ELEMENT ( "Bond Energy"          )
      WRITE ( PRINT_LINE, "(F16.4)" ) EANGLE    ; CALL PRINT_SUMMARY_ELEMENT ( "Angle Energy"         )
      WRITE ( PRINT_LINE, "(F16.4)" ) EDIHEDRAL ; CALL PRINT_SUMMARY_ELEMENT ( "Dihedral Energy"      )
      WRITE ( PRINT_LINE, "(F16.4)" ) EIMPROPER ; CALL PRINT_SUMMARY_ELEMENT ( "Improper Energy"      )
      WRITE ( PRINT_LINE, "(F16.4)" ) EELECT    ; CALL PRINT_SUMMARY_ELEMENT ( "Electrostatic Energy" )
      WRITE ( PRINT_LINE, "(F16.4)" ) ELJ       ; CALL PRINT_SUMMARY_ELEMENT ( "Lennard-Jones Energy" )
   END IF

   ! . Write out the constraint energy.
   WRITE ( PRINT_LINE, "(F16.4)" ) ECONSTRAINT ; CALL PRINT_SUMMARY_ELEMENT ( "Constraint Energy" )

   ! . Write out the terminator.
   CALL PRINT_SUMMARY_STOP

   END SUBROUTINE ENERGY_PRINT

END MODULE POTENTIAL_ENERGY
